//********************************************************************************
//*         Name:  coprocm.c                                                     *               
//*                                                                              *
//*       Author:  Daniel F. Ramirez                                             *
//*                                                                              *
//*          Date:  06/13/2003                                                   *
//*                                                                              *
//*   Description:                                                               *
//*                                                                              *
//*                                                                              *
//*     This is a Master Math Co-Processor application that uses a PIC18F8720 to * 
//* send expressions to the embedded Math Co-Processor using the SPI interface,  *
//* in order to compute floating point arithmetic expressions in RPN format.  It *
//* can be used for Sumo Robots that require floating point for navigation and   *
//* tracking of an oponent with minimal space, power and weight requirements are *
//* prime candidates for using this co-processor, since it is very compact.      *
//* The master can be a PIC or STAMP or any other micro-controller that requires *
//* a math co-processor and has at least 4 spare I/O pins. In addition to the    *
//* basic arithmetic functions, the math co-processor can also evaluate          *
//* expressions that contain trigonometric and scientific functions.             *
//*                                                                              *
//*      A communication interface is provided that allows the host to use high  *
//* speed asynchronous serial link or SPI or I2C Master to request floating point*
//* equation calculations or to pass it individual sensor readings, average      *
//* sensor readings, calibrating a sensor or to perform a sensor processing      *
//* algorithm using command message packets.  The commands available are         *
//* described below.  This header file provides the common Master and Slave      *
//* message formats and descriptors.                                             *
//*/*                                                                            *
//* The high priority ISR will invoke the "Z;" reset command via the serial      *
//* interface. Both the serial interface and the SPI interface are supported by  *
//* defining the approprate compiler flag.                                       *
//*                                                                              *
//*                                                                              *
//*     This master controller can be used to communicate with a PC host or any  *
//* USART, I2C or SPI slave, in order to  give the microcontroller the capability*
//* to serve as an SPI or I2C slave controller in order to provide it the        *
//* floating point coprocessor functions.                                        *  
//* It can also be used for robotic applications where multiple controllers and  *
//* SPI devices are used, with a minimum of data and control lines needed to     *
//* connect them.  Each I2C device has its own unique address and only listens   *
//* to commands associated with it.                                              *
//*                                                                              *
//*     Future upgrades to this application will include implementing various    *
//* filters including FIR, FFR, Kalman and Median filters.  Both FFT and IFFT    *
//* functions and a Histogram function are also in the works. Advanced algorithms*
//* such as Vector Force Fields (VFF), Fuzzy Logic and Neural Networks are new   *
//* areas that may be explored using this application.   In addition, this       *
//* application may be able to take advantage of the new DSPic Signal Processors *
//* (DSP's) become available from Microchip.                                     *
//*                                                                              *
//* Usage:                                                                       *
//*	The following Math Co-Processor Controller command messages are available:   *
//*                                                                              * 
//* calculate    - Evalueates the RPN expressions that are stored in the         *
//* message expression buffer.  Multiple expressions can be sent to the          *
//* co-processor for evaluation, if they are separated by a ";" semi-colon.      *
//* Large expressions can span multiple messages to the co-processor.            *
//*                                                                              * 
//* get_answers  - Get the results from the latest arithmetic operations         *
//* carried out during the last cycle. Up to 7 separate single precision floating*
//* point and 3 double precision  values may be returned during a single message *
//* transfer.                                                                    *
//*                                                                              *
//* clear_memory - Erases the contents of the Math co-processor's storage        * 
//* registers. These same registers are used for storing and retrieving temporary*
//* values. It clears floating point registers 0..MAX_FLOAT_REGISTERS.           *
//*                                                                              *
//* store   - Stores a floating point value to a selected floating point         *
//* register. The register ID range is 0..MAX_FLOAT_REGISTERS.                   *                                                            
//*                                                                              *
//* recall 	- Retrieves a floating point value from the selected floating        *
//* point register. The register ID range is 0..MAX_FLOAT_REGISTERS.             *                                                         
//*                                                                              *                                                                     
//*	GetStatus - Get the Math Co-processor status including current state, and    *  
//* any error conditions encountered i.e. division by 0, overflow, underflow.    *
//*                                                                              *
//* ResetCommand - Reset the DC motor Controller (brake the motor).  This        *
//* command can be used to implement an emergency stop function.                 *  
//*                                                                              *
//*                                                                              * 
//*                                                                              *
//********************************************************************************
    
 
// Include file headers here

#include "flags.h"                                      // Global flags used for MPSIM debugging 
  
#ifdef PIC18C252     
//#include <p18c252.h>					// Register definitions for PIC18C252
#endif

#ifdef PIC18C452     
#include <p18c452.h>					// Register definitions for PIC18F8720
#endif

#ifdef PIC18F452     
#include <p18f452.h>					// Register definitions for PIC18F8720
#endif
  
#ifdef PIC18F8720     
#include <p18f8720.h>
#endif
 
#ifdef PIC18F6620     
#include <p18f6620.h>
#endif
 
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <spi.h>                                        // SPI library functions
#include <pwm.h>                                        // PWM library functions
#include <timers.h>                                     // Timer library functions
#include <portb.h>                                      // PORTB library functions                                 
#include <usart.h>                                      // USART library functions
#include <delays.h>                                     // High resolution delay functions
//#include "portutil.h"                                 // Port utility functions
#include "timeutil.h"
#include "stdio.h"                                      // Serial I/O support functions
#include "network.h"                                   	// SPI Networking header
//#include "dcmotor.h"                                   	// SPI DC Motor Controller common message header
#include "spimast.h"                                   	// SPI Master services 
#include "coproc.h"										// Include Common Math Co-processor header
#include "coprocm.h"									// Include Math Co-processor Services header
//#include "dcmotorm.h"                                   // DC Motor Controller functions
  
//********************************************************************************
// Constant Definitions
//********************************************************************************
#define CR 13


//#define TEST                                          	// Flag used for debug I/O
//#define USART_TEST                                    // Test serial I/O with this application
//#define GARDENBOT                                     // Control a GardenBot robot
#define GEARBOT                                         // Control a GearBot

//#define SERIAL_INTERFACE								// Use high speed serial link
//#define I2C_INTERFACE									// Use I2C interface
//#define SPI_INTERFACE                                 // Use SPI interface

//#define MASTER_TEST_0                                 // Select the SPI Master Controller test #0
//#define MASTER_TEST_1                                 // Select the SPI Master Controller test #1
//#define MASTER_TEST_2                                 // Select the SPI Master Controller test #2
//#define MASTER_TEST_3                                 // Select the SPI Master Controller test #3
//#define MASTER_TEST_4                      

#define NUMBER_CYCLES 	100								// Number of PWM Cycles to be sent to H-Bridge

#define PAUSE_TIME      100	 					  		// Pause between PWM commands (Acceleration) 


#define StackSize		 5								// HP-45 stack (5 levels)
#define RegisterSize	10								// HP-45 Registers (1-9)

// Declare global constant definitions here

#define TRUE 1
#define FALSE 0

#define INPUT 1											 // Port input direction 
#define OUTPUT 0										 // Port output direction 


#ifdef PARALLEL_PORT_INTERFACE

// Parallel Port Interface

#define BUSY					 PORTBbits.RB0          // Slave BUSY Status Pin (1=Busy, 0=Finished)
#define BUSY_DIR				 TRISBbits.TRISB0       // Slave BUSY Status Pin Direction  
 
#define RD_BAR					 PORTEbits.RE0          // RD_BAR Control Pin 
#define WR_BAR                   PORTEbits.RE1          // WR_BAR Control Pin   
#define CS_BAR                   PORTEbits.RE2          // CS_BAR Control Pin  

#define RD_BAR_DIR				 TRISEbits.TRISE0       // RD_BAR Control Pin Direction
#define WR_BAR_DIR               TRISEbits.TRISE1       // WR_BAR Control Pin Direction 
#define CS_BAR_DIR               TRISEbits.TRISE2       // CS_BAR Control Pin Direction 

// Bit definitions for flag register
 
typedef struct slave_port_flags
{
        unsigned ERR:1;				// err  	equ	00h		;Error flag bit
        unsigned OUTRDY:1;			// OUTRDY	equ	01h		;Output data ready flag
        unsigned INFULL:1;			// INFULL	equ	02h		;Input data received flag
        unsigned SPARE1:1;
        unsigned SPARE2:1;
        unsigned SPARE3:1;
        unsigned SPARE4:1;
        unsigned SPARE5:1;
 
} SLAVE_PORT_FLAGS ;

// Register definitions

static volatile SLAVE_PORT_FLAGS FlagReg;			// FLAGREG	equ	20h		;Flag bit register
byte OutData;						// OUTDATA	equ	21h		;Output data
byte InData;						// INDATA	equ	22h		;Input data
int	Count = 0;						// COUNT	equ	23h		;Count of times output register read


#endif

// Variables

int Status = 0;											// Status return code

byte ControllerState; 		// State of selected slave controller, range (0..255) 
											// 		0 => NO_STATES		No States    
											//		1 => INITIALIZED	The selected slave contoller is Initialzed 
											//		2 => DONE 	 	 	The selected slave contoller is done processing current command
											// 		3 => BUSY 	 		The selected slave contoller is busy processing current command 
											// 		4 => ERROR 	 		The selected slave contoller encountered an error processing the current command 

byte ControllerErrorCode; 	// Error code returned by the selected slave controller, range (0..255) 
											// 		0 => NO_ERRORS 		No error codes  
											// 		1 => FAILURE 	 	Failed to complete command
											//  	2 => SUCCESS  	 	Completed command successfully 
 
byte ControllerDiagnosticCode; 	// Diagnostic message code returned by the selected slave controller, range (0..255) 
											// 		0 => NO_DIAGNOSTICS 	No Diagnostic messages
											//  	1 => DIAGNOSTIC 1  	Diagnostic message 2 
											// 		2 => DIAGNOSTIC 2 	Diagnostic message 1
 
// Declare SPI Master Buffer variables here

byte SPI_Rx_Buffer[SPI_BUFFER_SIZE];                  	// SPI Receive Buffer
byte SPI_Tx_Buffer[SPI_BUFFER_SIZE];                  	// SPI Transmit Buffer
byte SPI_Start_Buffer=0;                         		// Start pointer for ring buffer
byte SPI_End_Buffer = 0;								// SPI End Buffer Index

byte SPI_Rx_Flag = FALSE;                        		// Indicates that data was received into the ring buffer
word SPI_Timeout = 0;                            		// SPI I/O Timeout counter
byte SPI_Timeout_Flag;                           		// SPI Timeout flag
word SPI_Number_Characters = 0;						 	// SPI Number of characters in buffer
int SPI_Communications_Error;                    		// SPI I/O error flag
                                                 		// -1 = SPI Input Bufer overrun error
                                                 		// -2 = SPI Communications timeout error
                                                 		// -3 = SPI buffer empty error

int SPI_Interrupt_Count = 0;                   	// Number of MSSP interrupts
int SPI_Data_In = 0;								// Data byte read from the SPI Master
int SPI_Data_Out = 0;							// Data byte sent to the SPI Master
int SPI_Status;									// SPI Status returned from routines
int MessageIndex = 0;							// Message ID number


// Initialize the Math Co-processor Commands

word SelectedCommand = 0;

typedef union message_union_type
{
    CALCULATE_TYPE CalculateCommand;							// Calculate command message
	GET_ANSWERS_TYPE GetAnswersCommand;							// Get answers command message
	CLEAR_MEMORY_TYPE CLearMemoryCommand;						// Clear memory command message
	STORE_TYPE StoreCommand;									// Store command message
	RECALL_TYPE RecallCommand;                          		// Recall command message	
 	GET_STATUS_COMMAND_TYPE GetStatusCommand;					// Get status message
	RESET_COMMAND_TYPE ResetCommand;					 		// Reset command message

	COMMAND_MESSAGE_TYPE SelectedCommand;						// Selected command	
	MESSAGE_TYPE Message;										// Unformatted message
} MESSAGE_UNION_TYPE;

// Define the Math Co-processor Commands message
MESSAGE_UNION_TYPE CoProcessorMessage;
  
#ifdef SINGLE_PRECISION
float Registers[MAX_FLOAT_REGISTERS];
float Answers[MAX_ANSWERS];
#else
double Registers[MAX_FLOAT_REGISTERS];
double Answers[MAX_ANSWERS];
#endif

char af1[BufferSize];    // Holds ascii representation of an IEEE floating point number for testing

//unsigned char arraywr[] = {1,2,3,4,5,6,7,8,9,10,11,
//12,13,14,15,16,0};
extern unsigned char arraywr[];

//24C040/080/160 page write size

//unsigned char *wrptr = arraywr;
//unsigned char arrayrd[16];
//unsigned char *rdptr = arrayrd;
extern unsigned char *wrptr = arraywr;
extern unsigned char arrayrd[16];
extern unsigned char *rdptr = arrayrd;

//unsigned char var;
unsigned char TheCharacterRead;								// Character read
unsigned char TheCharacterSent;								// Character sent
extern int SPIStatus;												// Status return from SPI function call
 
 
// Calculator specific variables

//char Buffer[BufferSize];					// Floating point Buffer
char StackLabel[]={'X','Y','Z','T'};		// Stack register labels
double Stack[StackSize];					// HP-45 Stack 
double Registers[RegisterSize];				// HP-45 Registers 1-9
double X = 0.0;								// X register (Display)
double Y = 0.0;								// Y register
double LastX = 0.0;							// Last X register
byte RegisterNumber = 0;					// Register number 1..9
//int CharacterCount = 0;								// Character count
int BufferCount = 0;								// Buffer count
int MessageBufferIndex = 0;							// Message buffer index

byte KeyValue; 
extern char Ch;

//**********************************************************************************
// Define Global variables common to all controllers here
//**********************************************************************************


byte ControllerState; 		// State of selected slave controller, range (0..255) 
											// 		0 => NO_STATES		No States    
											//		1 => INITIALIZED	The selected slave contoller is Initialzed 
											//		2 => DONE 	 	 	The selected slave contoller is done processing current command
											// 		3 => BUSY 	 		The selected slave contoller is busy processing current command 
											// 		4 => ERROR 	 		The selected slave contoller encountered an error processing the current command 

byte ControllerErrorCode; 	// Error code returned by the selected slave controller, range (0..255) 
											// 		0 => NO_ERRORS 		No error codes  
											// 		1 => FAILURE 	 	Failed to complete command
											//  	2 => SUCCESS  	 	Completed command successfully 
 
byte ControllerDiagnosticCode; 	// Diagnostic message code returned by the selected slave controller, range (0..255) 
											// 		0 => NO_DIAGNOSTICS 	No Diagnostic messages
											//  	1 => DIAGNOSTIC 1  	Diagnostic message 2 
											// 		2 => DIAGNOSTIC 2 	Diagnostic message 1


// FUNCTION Prototypes

void main(void);
#ifdef PARALLEL_PORT_INTERFACE
void InitializeParallelPort(void);
#endif

// Application specific functions

void set_wren(void);
void busy_polling(void);
unsigned char status_read(void);
void status_write(unsigned char data);
void byte_write(unsigned char addhigh, unsigned char
addlow, unsigned char data);
void page_write(unsigned char addhigh, unsigned char
addlow, unsigned char *wrptr);
void array_read(unsigned char addhigh, unsigned char
addlow, unsigned char *rdptr,
unsigned char count);
unsigned char byte_read(unsigned char addhigh,
unsigned char addlow);


#ifdef MASTER_COPROCESSOR 
//******************************************************************************
//* main - Main application sends commands to the PIC Math Co-Processor board.
//******************************************************************************

void main(void)
{ 
 	  int i; 												// Local Loop Index
	  int j; 												// Local Loop Index

   	SetupTimers();                                     	// Setup timers and software
 
  	// Configure USART for 20 MHz, 9600 Baud, 8-Bits, No parity, Asynchronous
   
	InitializeUsart(); 

	// Initialize the 8-Bit Parallel Slave Port as a Master 

	InitializeParallelPort();

 #ifdef PARALLEL_PORT_INTERFACE
  printf("PIC18C452 8-Bit Slave Parallel Port Controller Master Application ... \r\n");
     	
  
	while(1)									// Loop
	{
	
 			// Wait until the slave is not busy

			while(BUSY);


			// Send message to Slave Parallel Port  	
			
     		printf("Master: Send message. ");

   			TRISD = 0x00;                                          	// PORT D all outputs (switches select each pin state)

 			for (i=0; i<SPI_BUFFER_SIZE; i++)
			{
				SPI_Tx_Buffer[i] = i;

				// Toggle Control bits for write  
				CS_BAR = 0;
				WR_BAR = 0;
 
				// Write to the slave port
				PORTD = SPI_Tx_Buffer[i];

				// Toggle Control bits for write  
				CS_BAR = 1;
				WR_BAR = 1;
 

				printf("Master: Transmit i = ");
				dec(i);
				printf("  SPI_Tx_Buffer = ");
				dec(SPI_Tx_Buffer[i]);
 			
				printf("\r\n");
			
				//pause(1000);
			}
 			
			// Wait until the slave is not busy

			while(BUSY);

      		printf("Master: Receive message. ");
			//pause(1);

  			TRISD = 0xFF;                                          	// PORT D all inputs (switches select each pin state)
  		
			for (i=0; i<SPI_BUFFER_SIZE; i++)
			{
				// Toggle Control bits for read  
				CS_BAR = 0;
				//WR_BAR = 1;
				RD_BAR = 0;

 				// Read the slave port
				SPI_Rx_Buffer[i] = PORTD;

				// Toggle Control bits for read  
				CS_BAR = 1;
				//WR_BAR = 1;
				RD_BAR = 1;

				printf("Master: Receive: i = ");
				dec(i);			
				printf("  SPI_Rx_Buffer = ");
				dec(SPI_Rx_Buffer[i]);
			
				printf("\r\n");
			
				//pause(1000);
 
 
			}

		}

#endif


#ifdef SPI_BUFFER_INTERFACE
  	 // Initialize the SPI Master hardware
   
   	InitializeSPIMaster();
#endif

#ifdef TEST	
	// No debug I/O allowed when using the USART interface because it interferes with the messages to the co-processor

 	printf("Master Math Co-Processor Application ...\n");
/* 		
 	printf("Delay 5 seconds ...\r\n");
 	pause(5000);
 	printf("End of 5 second delay. \r\n");
*/
#endif
   


   	// Test #1
   	// Send lower case characters to the SPI Co-Processor and read them back
   	// the SPI Co-Processor has converted them to upper case.
   	
 	TheCharacterSent = 0;
 	MessageIndex = 0;
 	
 	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
 		CoProcessorMessage.Message.Data[i] = 0;            			// Clear the message  
 	}

    // Test code
    
 	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
 		SPI_Rx_Buffer[i] = 0;                  		// Clear SPI Receive Buffer 
 		//SPI_Tx_Buffer[i] = i;                  	// Initialize SPI Transmit Buffer to a ramp 
 		SPI_Tx_Buffer[i] = i;      	// Initialize SPI Transmit Buffer to a ramp 
	}
 
    
 
    
 
	//SPI_CS = ENABLE; 								// Assert chip select
    //SPI_CS = DISABLE; 							// Negate chip select
    
    SelectedCommand = 0;

 
		// Mainline Logic for SPI Buffer Interface goes here ...
		
		// This is the section where user selected command messages are issued to the Math Co-Processor 
		// In this case High Level Math Co-ProceController Commands are issued here ....
		// These functions may be used to test all the available the Math Functions and evaluate algorithms.			
 

	//********************************************************************************************
 	//* Send various arithmetic expressions to the Math Co-processor for evaluation
	//********************************************************************************************

    while (1)
    {
  		
 		 // Send the Calculate command message that evaluates RPN expression...
 		
	    // Send Calculate Command
		SendCalculateCommand ();
 					
  		// Send Get Answers command message in order to return the floating point answers to the master
  		SendGetAnswers();
 
  		// Send Get Status Command();
 		SendGetStatusCommand();
		
#ifdef NEW_CODE
  			
  		// Send the Clear Memory Command message that erases the floating point registers
		SendClearMemoryCommand();
					
 		// Send the Store Command message that saves a floating point value to a selected float register
		SendStoreCommand();
					
  		// Send the Recall Command message that  Retrieves a floating point value from a 
  		// selected float register.	
  		SendRecallCommand();
   
   		// Send the Filter command message that performs selected filter 
  		// operation (FIR, IFR)

		SendFilterCommand();
 				
   		// Send the FFT command message that performs the selected 
  		//FFT operation on the specified data.

		SendFFTCommand();
 				
   		// Send the Inverse Filter command message that performs an inverse
  		// FFT operation on the specified data.

		SendIFFTCommand();
 				
   		// Send the Kalman Filter command message that performs a Kalman
  		// Filter on the specified data

		SendKalmanFilterCommand();
 				
   		// Send the Median Filter command message that performs a Median filter 
  		//operation on the specified data.

		SendMedianFilterCommand();
 				
   		// Send the Histogram command messagePerforms a Histogram operation

		SendHistogramCommand();
 				
   		// Process the Sort command message that performs the selected Sort operation 
  		//on the specified data

		SendSortCommand();
 
		
 		//**********************************************************************************
 		// Send and process the Get Status command response message from the Slave
		//**********************************************************************************
 
 		SendGetStatusCommand();
  	
 		//**********************************************************************************
 		// Send and process the Reset Command response message from the Slave
		//**********************************************************************************
 
 		//SendResetCommand();
#endif	
			
	}
		
#endif
	
	


	// close the spi port

	CloseSPI();
	
 /*	
 	// 24LC256 I2C Serial EEPROM Test
	set_wren();
	status_write(0);
	busy_polling();
	set_wren();
    byte_write(0x00, 0x61, 'E');
	busy_polling();
	var = byte_read(0x00, 0x61);
	set_wren();

 
	page_write(0x00, 0x30, wrptr);
	busy_polling();
	array_read(0x00, 0x30, rdptr, 16);
	var = status_read();
  */
	// Done.

	while(1);
}
 
//**********************************************************************************
//* SendSPIMessage - Sends a command message to the SPI Slave and stores the data received 
//* into the SPI_Rx_Buffer so that it may be processed by Reveive Message. 
//**********************************************************************************
byte SendSPIMessage(void)
{
    int i;														// Loop index 		
	byte SendSPIMessageStatus = TRUE;								// Status return

 
#ifdef TEST
 
	// Send a test ramp that spans 0..SPI_BUFFER_SIZE
	
	for (i=0; i<DATA_BUFFER_SIZE; i++)
	{
		CoProcessorMessage.Message.Data[i] = i;
  	}	
  	 	
#endif	
	
	// Insert SYNC byte into each message
	
	SPI_Tx_Buffer[0] =  SYNC_BYTE;
	SPI_Tx_Buffer[1] =  SYNC_BYTE;
	
 	// Copy the Ramp test message from the SPI Transmit Buffer to the 
 	// Motor Message for test
    			
  	//memcpy( SPI_Tx_Buffer+TX_DATA_OFFSET, &CoProcessorMessage.Message.Data, DATA_BUFFER_SIZE);  

   	for (i=2; i<DATA_BUFFER_SIZE; i++)
	{
  			 
		SPI_Tx_Buffer[i] =  CoProcessorMessage.Message.Data[i];
	}
	
	//**********************************************************************************
	// Send an SPI message command here
	//**********************************************************************************
/*  
	SPI_CS = DISABLE; 					// De-assert chip select
    pause(1); 							// Wait some time to allow it to be processed.
	SPI_CS = ENABLE; 					// Assert chip select
    pause(1); 							// Wait some time to allow it to be processed.
*/	
 	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
   	   		
         											// We may want handshaking here..
#ifdef USART_BUFFER_INTERFACE
     		
		//Write1USART(SPI_Tx_Buffer[i]);	// Send a character to the  Slave Co-Processor	
		//send_byte(SPI_Tx_Buffer[i]);
		hex2(SPI_Tx_Buffer[i]);
        send_byte(CR);
      	pause(1); 							// Wait some time to allow it to be processed.
  
#endif
        										
#ifdef SPI_BUFFER_INTERFACE
        //SPIStatus = putcSPI(MotorMessage.Message.Data[i]);	// Send a character to the  Slave Co-Processor
        SPIStatus = putcSPI(SPI_Tx_Buffer[i]);	// Send a character to the  Slave Co-Processor
#endif



#ifdef USART_BUFFER_INTERFACE
       	// Read a single byte from the USART into the serial 
       	// receiver buffer.
	
		//SPI_Rx_Buffer[SPI_End_Buffer] = Read1USART();	
		//SPI_Rx_Buffer[SPI_End_Buffer] = get_byte();
#endif

#ifdef SPI_BUFFER_INTERFACE
		// Read the data sent from the SPI Slave and store it in the SPI receiver buffer
		TheCharacterRead = ReadSPIByte(); 
		SPI_Rx_Buffer[i] = TheCharacterRead;
#endif

 
 
 	}
 	
    //SPI_CS = DISABLE; 						// Negate chip select
    //pause(10); 							// Wait some time to allow it to be processed.
	
#ifdef TEST		
	

	// Display contents of SPI Buffers 	
			 
	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
		printf("Master: i = ");
		dec(i);
		printf("  SPI_Tx_Buffer = ");
		dec(SPI_Tx_Buffer[i]);
			
		printf("  SPI_Rx_Buffer = ");
		dec(SPI_Rx_Buffer[i]);
        printf("  ");
        send_byte(SPI_Rx_Buffer[i]);
			
		printf("\r\n");
			
		//pause(1000);
	}
	
	printf("\r\n\r\n");
/*	

	// Display contents of SPI message
	
	for (i=0; i<DATA_BUFFER_SIZE; i++)
	{
		printf("Master: i = ");
		dec(i);
		printf("  MotorMessage.Message.Data = ");
		dec(MotorMessage.Message.Data[i]);
		printf("\r\n");
	}
*/

#endif
		
	return SendSPIMessageStatus;
}

//**********************************************************************************
//* ReceiveSPIMessage - Receives and processes the message that is in the SPI_Rx_Buffer 
//* which was simultaneously sent back from the SPI Slave Controller.  The message may
//* contain data or status information.
//**********************************************************************************
byte ReceiveSPIMessage(void)
{
	int i; 															// Loop index 
 	byte ReceiveSPIMessageStatus = TRUE;							// Status return
/*		 				
  	// Copy the entire DC Motor Controller message received from the SPI Slave
   	// to local storage for processing.  It may include sensor data, telemetry and status 
   	// information and is received concurrently with each message that the Master SPI controller sends.
   			
  	//memcpy(&CoProcessorMessage.Message.Data, (char *) (SPI_Rx_Buffer+RX_DATA_OFFSET), DATA_BUFFER_SIZE);  
  	
   			
   	for (i=0; i<DATA_BUFFER_SIZE; i++)
	{
  			
		CoProcessorMessage.Message.Data[i] = SPI_Rx_Buffer[i];
	}
     		
*/
	
	//**********************************************************************************
	// Receive an SPI message from the Math Co-processor and process it  here
	//**********************************************************************************
/*  
	SPI_CS = DISABLE; 					// De-assert chip select
    pause(1); 							// Wait some time to allow it to be processed.
	SPI_CS = ENABLE; 					// Assert chip select
    pause(1); 							// Wait some time to allow it to be processed.
*/	
 	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
   	   		
        //pause(1); 							// Wait some time to allow it to be processed.
         											// We may want handshaking here..
#ifdef USART_BUFFER_INTERFACE

       	// Send dummy message to the co-processor 
         										
  
 		//send_byte(SPI_Tx_Buffer[i]);	// Send a character to the SPI Slave Co-Processor	
    
		if (DataRdyUSART())
    	{
			// Read the data sent message sent from the Math Co-processor and store it in the receiver buffer
 	
			SPI_Rx_Buffer[SPI_End_Buffer] = ReadUSART();		
		}		
 
#endif

#ifdef SPI_BUFFER_INTERFACE
    
       	// Send dummy message to the co-processor 
         										
        //SPIStatus = putcSPI(CoProcessorMessage.Message.Data[i]);	// Send a character to the SPI Slave Co-Processor
        SPIStatus = putcSPI(SPI_Tx_Buffer[i]);	// Send a character to the SPI Slave Co-Processor

		
 		if (DataRdySPI())
		{         
			// Read the data sent from the SPI Math Co-processor and store it in the SPI receiver buffer
 
   	   		TheCharacterRead = ReadSPIByte();  
   	   		//pause(1);	   		
  			SPI_Rx_Buffer[i] = TheCharacterRead;
		}
#endif


 
 	}
 	
    //SPI_CS = DISABLE; 						// Negate chip select
    //pause(10); 							// Wait some time to allow it to be processed.
 
 
	
#ifdef TEST	
	

	// Display contents of SPI Buffers 	
			 
	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
		printf("Master: i = ");
		dec(i);
		printf("  SPI_Tx_Buffer = ");
		dec(SPI_Tx_Buffer[i]);
			
		printf("  SPI_Rx_Buffer = ");
		dec(SPI_Rx_Buffer[i]);
			
		printf("\r\n");
			
		//pause(1000);
	}
	
	printf("\r\n\r\n");
/*	

	// Display contents of SPI message
	
	for (i=0; i<DATA_BUFFER_SIZE; i++)
	{
		printf("Master: i = ");
		dec(i);
		printf("  CoProcessorMessage.Message.Data = ");
		dec(CoProcessorMessage.Message.Data[i]);
		printf("\r\n");
	}
*/

#endif

 	 				
  	// Copy the entire message received from the math coprocessor 
   	// to local storage for processing.  It may include sensor data, telemetry and status 
   	// information and is received concurrently with each message that the Master SPI controller sends.
   			
  	//memcpy(&CoProcessorMessage.Message.Data, SPI_Rx_Buffer+RX_DATA_OFFSET, DATA_BUFFER_SIZE);  
  			  
  			
   	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
  			
		CoProcessorMessage.Message.Data[i] = SPI_Rx_Buffer[i];
	}
    		
 	//**********************************************************************************************
 	// Display the message received from the SPI Slave here
 
 
    // Process the message received  from the Math Co-processor Slave here ...
			        	
 	switch (CoProcessorMessage.SelectedCommand.Command)
 	{
		case CALCULATE_COMMAND:
		{
 			// Process the Calculate command message that evaluates RPN expression
 			ReceiveCalculateCommand();
 				
			break;
		} 
			
		case GET_ANSWERS_COMMAND:
		{
		
 			// Process Get Answers command message in order to return the floating point answers to the master
  			ReceiveGetAnswers();
  					
  			break;
		}
			
		case CLEAR_MEMORY_COMMAND:
		{
  			// Process the Clear Memory Command message that erases the floating point registers
			ReceiveClearMemoryCommand();
					
 			break;
		}
			
		case STORE_COMMAND:
		{

 			// Process the Store Command message that saves a floating point value to a selected float register
			ReceiveStoreCommand();
					
   			break;
		}
			
		case RECALL_COMMAND:
		{

  			// Process the Recall Command message that  Retrieves a floating point value from a 
  			// selected float register.	
  			ReceiveRecallCommand();
   			break;
		}
#ifdef NEW_CODE
		case FILTER_COMMAND:
		{
  			// Process the Filter command message that performs selected filter 
  			// operation (FIR, IFR)

			ReceiveFilterCommand();
    		break;
		}
				
		case FFT_COMMAND:
		{
  			// Process the FFT command message that performs the selected 
  			//FFT operation on the specified data.

			ReceiveFFTCommand();
    		break;
		}
				
		case IFFT_COMMAND:
		{
  			// Process the Inverse Filter command message that performs an inverse
  			// FFT operation on the specified data.

			ReceiveIFFTCommand();
    		break;
		}
				
		case KALMAN_FILTER_COMMAND:
		{
  			// Process the Kalman Filter command message that performs a Kalman
  			// Filter on the specified data

			ReceiveKalmanFilterCommand();
    		break;
		}
				
		case MEDIAN_FILTER_COMMAND:
		{
  			// Process the Median Filter command message that performs a Median filter 
  			//operation on the specified data.

			ReceiveMedianFilterCommand();
    		break;
		}
				
		case HISTOGRAM_COMMAND:
		{	
  			// Process the Histogram command messagePerforms a Histogram operation

			ReceiveHistogramCommand();
    		break;
		}
				
		case SORT_COMMAND:
		{
  			// Process the Sort command message that performs the selected Sort operation 
  			//on the specified data

			ReceiveSortCommand();
    		break;
		}
#endif			
		case GET_STATUS_COMMAND:
		{
			//**********************************************************************************
 			// Receive and process the Get Status command response message from the Slave
			//**********************************************************************************
 
 			ReceiveGetStatusResponse();
   			break;
		}
 	
		case RESET_COMMAND:
		{
			//**********************************************************************************
 			// Receive and process the Reset Command response message from the Slave
			//**********************************************************************************
 
 			ReceiveResetResponse();
   			break;
		}	

	}
		
	
 	//**********************************************************************************************
	
	
	//pause(2000);
	 	
 	// Generate next command 
		
	//SelectedCommand++;   
	//SelectedCommand = SelectedCommand % 6;
 
    
#ifdef SPI_NETWORK_INTERFACE
	// Mainline Logic for SPI Network Interface goes here ...
		
	SPI_CS = ENABLE; 						// Assert chip select
 
    putsSPI(SPI_Tx_Buffer);					// Send the message stored in the SPI_Tx_Buffer to the 
                                  				// SPI Slave Co-Processor
 
    //pause(5); 								// Wait some time to allow it to be processed.
         										// We may want handshaking here...
 
    // Read the message data sent from the SPI Slave
   	getsSPI(SPI_Rx_Buffer, sizeof(SPI_Rx_Buffer));
 
 	
/*   	
   	// Display the data sent and received
   	   	 
	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
		printf("Master: i = ");
		dec(i);
		printf("  SPI_Tx_Buffer = ");
		dec(SPI_Tx_Buffer[i]);
		
		printf("  SPI_Rx_Buffer = ");
		dec(SPI_Rx_Buffer[i]);
			
		printf("\r\n");
			
		//pause(1000);
  
	}
*/
   	   	
    SPI_CS = DISABLE; 						// Negate chip select
  
#endif

   return ReceiveSPIMessageStatus;
}
 	
//**********************************************************************************
//* CheckControllerState - Check the status of the the current controller to see if
//* the command has been executed correctly.  Check the following Controller state
//* variables: 
//*		ControllerState	{ NO_STATES=0, INITIALIZED=1, DONE=2, BUSY=3, ERROR=4 }
//* 	ControllerErrorCode  { NO_ERRORS=0, FAILURE=1, SUCCESS=2 }	 
//* 	ControllerDiagnosticCode { NO_DIAGNOSTICS=0, DIAGNOSTIC1=1, DIAGNOSTIC2=2 }    
//**********************************************************************************
 byte CheckControllerState(void)
 {

	byte CheckControllerStateStatus = NO_STATES;
	
	// Start out with Controller State set to "NO_STATES" and continue to send and
	// receive Get Status messages until either the controller state changes to "DONE" or
	// gets an error state.
	
    ControllerState =  NO_STATES;
    
	SendGetStatusCommand();
	ReceiveGetStatusResponse(); 
	
	while (ControllerState != DONE)
	{
		  SendGetStatusCommand();
		  ReceiveGetStatusResponse();
#ifdef TEST	  
		  printf("ControllerState = ");
		  dec(ControllerState);
		  printf("  ControllerErrorCode = ");
		  dec(ControllerErrorCode);
		  printf("  ControllerDiagnosticCode = ");
		  dec(ControllerDiagnosticCode);
		  printf("\r\n");
		  
#endif	
	  
		  CheckControllerStateStatus = ControllerState;
		  if (ControllerState == ERROR)
		  {	  
		  		printf("Error # ");
		  		dec(ControllerErrorCode);
		  		printf(" was encountered by the Math Co-processor!!\r\n");
		  		
		  		break;
		  }
	}
			

#ifdef TEST	  
	if (ControllerState == DONE)
	{
		printf("Math Co-processor Command completed successfully!\r\n");
	}
#endif 
	return CheckControllerStateStatus;
	
}

//**********************************************************************************
//**********************************************************************************
//**********************************************************************************
//**********************************************************************************

// Send command messages to Math Co-processor here...
	 		
//**********************************************************************************
//* SendCalculateCommand - Send the Calculate commant to the Math Co-processor so  * 
//* that it can Evalueate the RPN expressions that are stored in the message       *
//* expression buffer. Return a status byte indicating the outcome of the Calculate*
//* Command process.                                                               *
//**********************************************************************************
byte SendCalculateCommand(void)  
{
 	byte CalculateCommandStatus = TRUE;							// Status return
  	word CalculateCommandDelay = 2;								// Delay to allow the duty command to complete
	byte i;														// Local loop index
	
 	// Send a message to the DC Motor Controller to start the DC Motor Calibration 
 	// process for the selected motor  ...

 	// Clear the SPI send and receive buffers
	//FlushSPIBuffers();
		
	// Build the Calculate Command 

 

	// Test #1 - Send up to 7 mathematical expressions to be evaluated by the 
	// Math Co-Processor.
 	// Send the Calculate command message that evaluates RPN expression...
 		
/*

// General Math Coprocessor functions

#define GOLD    'a'					// Gold function key    
#define ENTER	'b'                 // Enter key  
#define ROLL    'c'					// Roll key   
#define SWAP    'd'					// Swap X with Y key  
#define STORE   'e'					// Store X into register N, N=1..9   
#define RECALL  'f'					// Recall register N, N=1..9 into X  
#define LASTX   'g'                 // Recall register Last X into X

// Math Operators 

#define PLUS	'+'					// "+" Operator
#define MINUS	'-'					// "-" Operator
#define TIMES	'*'					// "*" Operator
#define DIVIDE	'/'					// "/" Operator

// Math functions

// Trigonometric Functions

#define SIN		'A'					// Sine function
#define ASIN  	'B'					// Tangent function
#define COS		'C'					// Cosine function
#define ACOS  	'D'					// Arc Cosine function
#define TAN		'E'					// Tangent function
#define ATAN	'F'					// Arc Tangent function
#define ATAN2	'G'					// Arc Tangent Two function
 
// Scientific Functions

#define SQRT	'H'					// Square root function
#define SQR		'I'					// Square function
#define EXP		'J'					// Exponential function E**X
#define LOG		'K' 				// Log function
#define LOG10	'L' 				// Log base 10 function
#define POWER	'M' 				// X**N  function
#define MAX		'N'                 // Maximum of x and y
#define MIN		'O'                 // Minimum of x and y
#define ABS		'P'                 // Absolute value of x
#define SIGN	'Q'                 // Sign of x and y

// Special functions

#define RAD 	'R'					// Convert to radians
#define DEG 	'S'					// Convert to degrees
#define FIX 	'T'					// Rounds x towards 0
#define FLOOR 	'U'					// Rounds x towards - Infinity
#define CEIL 	'V'					// Rounds x towards + Infinity
#define ROUND 	'W'					// Round x to nearest integer
#define PRECISION 	'X'				// Specify maximum number of digits precision up to machine precision used for display purposes
#define MOD		'Y'					// Mod Function identical to C mod function 
*/

   	// Test data for the math coprocessor application "b" is the "ENTER" key  
   
 		
 	// This is how the buffer is initialized for evaluating an expression from an embedded application...
	//strcpy(&CoProcessorMessage.CalculateCommand.Buffer, (char *)"25.0Hb3*b2+;\0");		// Test #1 - Calculate the expression: sqrt(25.0)*3.0 + 2  = 17.0 
	//strcpy(&CoProcessorMessage.CalculateCommand.Buffer, "30.0RA;\0");		// Test #2 - Calculate the expression: sin(radians(30)) =
	//strcpy(&CoProcessorMessage.CalculateCommand.Buffer, "3b4+8b6+/\0");	// Test #3 - Calculate the expression: (3+4)/(8+6) = 0.5
	//strcpy(&CoProcessorMessage.CalculateCommand.Buffer, "4.0b3.0b2.0b1.0b5.0b8.0+\0");	// Test #4 - Calculate the expression: 
	//strcpy(&CoProcessorMessage.CalculateCommand.Buffer, "3.0b5*b4.0b6/+\0");	// Test #5 - Calculate the expression: (3 * 5) + (4 /6)
	//strcpy(&CoProcessorMessage.CalculateCommand.Buffer, "8.0b2.0-b6.0b3.0+*\0");	// Test #5 - Calculate the expression: (8-2)*(6+3)

 
 	
	CoProcessorMessage.CalculateCommand.MessageID = MessageIndex++;		// Message ID range (0000 - FFFF) Hex
	CoProcessorMessage.CalculateCommand.Command = CALCULATE_COMMAND;	// Calculate command evaluates RPN expression

	// Test #1 - Calculate the expression: sqrt(25.0)*3.0 + 2  = 17.0 
	CoProcessorMessage.CalculateCommand.Buffer[0] = '2';
	CoProcessorMessage.CalculateCommand.Buffer[1] = '5';
	CoProcessorMessage.CalculateCommand.Buffer[2] = '.';
	CoProcessorMessage.CalculateCommand.Buffer[3] = '0';
	CoProcessorMessage.CalculateCommand.Buffer[4] = 'H';
 	CoProcessorMessage.CalculateCommand.Buffer[5] = 'b';
 	CoProcessorMessage.CalculateCommand.Buffer[6] = '3';
	CoProcessorMessage.CalculateCommand.Buffer[7] = '*';
	CoProcessorMessage.CalculateCommand.Buffer[8] = 'b'; 
	CoProcessorMessage.CalculateCommand.Buffer[9] = '2';		  
	CoProcessorMessage.CalculateCommand.Buffer[10] = '+';
    CoProcessorMessage.CalculateCommand.Buffer[11] = ';';
 	CoProcessorMessage.CalculateCommand.Buffer[12] = '\0';	  
	
	// Initialize the message buffer with an RPN expression to be evaluated by the Math co-processor 
	// ???????????????
/*	
	// *****************Missing logic to enter the expression from keyboard or directly from algorithm *****************
	for (i=0; i<MESSAGE_BUFFER_SIZE;i++)
	{
		CoProcessorMessage.CalculateCommand.Buffer[i] = 0;
	}
*/	
	
#ifdef TEST		

    printf("Master: Send Calculate Command  Command...\r\n");
  	printf("MessageID = ");
   	dec(CoProcessorMessage.CalculateCommand.MessageID);
   	printf(",  Buffer = !");
   	for (i=0; i<20;i++)
   	{
   		printf("i = ");
   		dec(i);
   		printf("Buffer = ");
   		dec(CoProcessorMessage.CalculateCommand.Buffer[i]);
   		printf("\r\n");
   	}
   	//string(CoProcessorMessage.CalculateCommand.Buffer);
   	printf("! \r\n\r\n");
#endif 

	// Send the command message to the SPI Slave Controller
	SendSPIMessage();
	
	// Receive and process the response message from the SPI Slave Controller
	ReceiveSPIMessage();
	
	// Give enough time for the command to complete 
	// ???????????????
	
	pause(CalculateCommandDelay);
	
 	// Check the status of the the current controller to see if the command has been 
	// executed correctly.
			
	//while (CheckControllerState() != DONE);
 
	return CalculateCommandStatus;
} 

//**********************************************************************************
//* SendGetAnswers  - Send command to the Math Co-procesor to get the results from *
//* the latest arithmetic operations carried out during the last cycle. Up to 7    *
//* separate single precision floating point and 3 double precision  values may be *
//* returned during a single message transfer.  Return a status byte indicating the*
//* outcome of the Send Get Answers command.                                       *
//**********************************************************************************
 
byte SendGetAnswers(void)
{
 	byte GetAnswersStatus = TRUE;								// Status return
 	word GetAnswersDelay = 20;									// Delay to allow the duty command to complete
 	byte i;														// Local loop index
 	
	// Clear the SPI send and receive buffers
	//FlushSPIBuffers();
		
	// Build the Get Answers Command message 
			
	CoProcessorMessage.GetAnswersCommand.MessageID = MessageIndex++;				// Message ID range (0000 - FFFF) Hex 
  	CoProcessorMessage.GetAnswersCommand.Command = GET_ANSWERS_COMMAND;  			// Gets the floating point answers (see enumeration
               			                 									        // declaration above.)                                 	 
	CoProcessorMessage.GetAnswersCommand.NumberOfAnswers = 0;		
 
#ifdef TEST
    printf("Master: Sending Get Answers Command: ");
  	printf("  Message ID = ");
    dec(CoProcessorMessage.GetAnswersCommand.MessageID);
    printf("\r\n\r\n");
#endif


	// Send the command message to the SPI Slave Controller to execute the command
	SendSPIMessage();
	pause(GetAnswersDelay);
	// Send the command message a second time to the SPI Slave Controller for the response
	SendSPIMessage();
	
	// Receive and process the response message from the SPI Slave Controller
	ReceiveSPIMessage();
	
	// Give enough time for the command to complete
	// ???????????????
	
	pause(GetAnswersDelay);
	
 	// Check the status of the the current controller to see if the command has been 
	// executed correctly.
			
	//while (CheckControllerState() != DONE);
 	
	return GetAnswersStatus;
}
			
//**********************************************************************************
//* SendClearMemoryCommand - Send the clear memory command that erases the contents*
//* of the Math co-processor's storage registers. These same registers are used for*
//* storing and retrieving temporary values. It clears floating point registers    *
//* 0..MAX_FLOAT_REGISTERS. Return a status byte indicating the outcome of the     *
//* send clear memory command.                                                     *
//**********************************************************************************
byte SendClearMemoryCommand(void)
{
 	byte ClearMemoryCommandStatus = TRUE;										// Status return
 	word ClearMemoryCommandDelay = 2;											// Delay to allow the duty command to complete

 	// Clear the SPI send and receive buffers
	//FlushSPIBuffers();
 				
  	// Build the Clear Memory Command message
 	
	CoProcessorMessage.CLearMemoryCommand.MessageID = MessageIndex++; 			// Message ID (0000 - FFFF) Hex
	CoProcessorMessage.CLearMemoryCommand.Command = CLEAR_MEMORY_COMMAND;		// Clears the floating point scratch registers (see enumeration
        		                        										// declaration above.)  
 
#ifdef TEST		

    printf("Master: Send Clear Memory Command...\r\n");
  	printf("MessageID = ");
    dec(CoProcessorMessage.CLearMemoryCommand.MessageID);
    printf("  Command = CLEAR_MEMORY_COMMAND");
    printf("\r\n\r\n");
#endif 
	
	// Send the command message to the SPI Slave Controller
	SendSPIMessage();
	
	// Receive and process the response message from the SPI Slave Controller
	ReceiveSPIMessage();
	
	// Give enough time for the command to complete
	
	pause(ClearMemoryCommandDelay);
	//??????????????????????????
	
 	// Check the status of the the current controller to see if the command has been 
	// executed correctly.
			
	//while (CheckControllerState() != DONE);
 
	return ClearMemoryCommandStatus;
}

//**********************************************************************************
//* SendStoreCommand - Sends store command that stores a floating point value to a *
//* selected floating point register.     The register ID range is                 *             
//* 0..MAX_FLOAT_REGISTERS.                                                        *
//**********************************************************************************
byte SendStoreCommand(void)
{
 	byte StoreCommandStatus = TRUE;							// Status return
 	word StoreCommandDelay = 2;								// Delay to allow the Store command to complete

 	// Clear the SPI send and receive buffers
	//FlushSPIBuffers();
 
 	// Build Store Command message
 		
	CoProcessorMessage.StoreCommand.MessageID = MessageIndex++; 	// Message ID (0000 - FFFF) Hex

  	CoProcessorMessage.StoreCommand.Command = STORE_COMMAND;		// Stores a floating point value to a selected float register
  
  	     			                           						// (see enumeration declaration above.)  
 		
	// *****************Missing logic to enter the register ID and value
	// from keyboard or directly from algorithm *****************
	// ???????????????

 	// Test #2 - Send up to MAX_FLOAT_REGISTER values to be stored in the Math 
 	// Co-processor's memory. First, initialize the message register ID and 
 	// register value with test data for input to the Math Co-processor
 
 	CoProcessorMessage.StoreCommand.RegisterID = 0; 			// Selected register range (0..MAX_FLOAT_REGISTERS)  
 	CoProcessorMessage.StoreCommand.RegisterValue = 3.14159263; // Selected register range (0..MAX_FLOAT_REGISTERS)  
 	//CoProcessorMessage.StoreCommand.RegisterID = 1; 			// Selected register range (0..MAX_FLOAT_REGISTERS)  
 	//CoProcessorMessage.StoreCommand.RegisterValue = 2.781828; // Selected register range (0..MAX_FLOAT_REGISTERS)  

#ifdef TEST		

    printf("Master: Store Command...\r\n");
  	printf("MessageID = ");
    dec(CoProcessorMessage.StoreCommand.MessageID );
    printf("  Register ID = ");
    dec(CoProcessorMessage.StoreCommand.RegisterID);   
    printf(",  Register Value = ");
    dec(CoProcessorMessage.StoreCommand.RegisterValue);
	DoubleToAscii(CoProcessorMessage.StoreCommand.RegisterValue, af1);
   	puts(af1);
    printf("\r\n\r\n");
#endif	
		
	// Send the command message to the SPI Slave Controller
	SendSPIMessage();
	
	// Receive and process the response message from the SPI Slave Controller
	ReceiveSPIMessage();
	
	// Give enough time for the command to complete
	
	pause(StoreCommandDelay);
	//???????????????????????????????
		
 	// Check the status of the the current controller to see if the command has been 
	// executed correctly.
			
	//while (CheckControllerState() != DONE);
 
	return StoreCommandStatus;
}
  					
//**********************************************************************************
//* SendRecallCommand - Sends the recall command that retrieves a floating point   *
//* value from the selected floating point register. The register ID range is      *
//* 0..MAX_FLOAT_REGISTERS.                                                        *
//**********************************************************************************
byte SendRecallCommand(void)
{
 	byte RecallCommandStatus = TRUE;									// Status return
 	word RecallCommandDelay = 2;										// Delay to allow the duty command to complete

 	// Clear the SPI send and receive buffers
	//FlushSPIBuffers();
 				
  	// Build Recall Command message
 
 	CoProcessorMessage.RecallCommand.MessageID = MessageIndex++;   		// Message ID (0000 - FFFF) Hex

  	CoProcessorMessage.RecallCommand.Command = 	RECALL_COMMAND;  		// Retrieves a floating point value from a selected float register	
   	                                							    	// Command (see enumeration declaration above.)  
	// *****************Missing logic to enter the register ID and value
	// from keyboard or directly from algorithm *****************
   	//CoProcessorMessage.RecallCommand.RegisterID = 0;
	// ???????????????

 	// Test #3 - Recall up to MAX_FLOAT_REGISTERS values from the Math 
	// Co-processors memory.
	
 	// Initialize the message register ID with test data 
 	// for input to the Math Co-processor
 
 	CoProcessorMessage.RecallCommand.RegisterID = 0;
 	//CoProcessorMessage.RecallCommand.RegisterID = 1;
	//CoProcessorMessage.RecallCommand.RegisterID = 2;
	//CoProcessorMessage.RecallCommand.RegisterID = 3;
	//CoProcessorMessage.RecallCommand.RegisterID = 4;
	//CoProcessorMessage.RecallCommand.RegisterID = 5;
	//CoProcessorMessage.RecallCommand.RegisterID = 6;
  					
#ifdef TEST		

    printf("Master: Send Recall Command...\r\n");
  	printf("MessageID = ");
  	dec(CoProcessorMessage.RecallCommand.MessageID);
    printf("  Register ID = ");
    dec(CoProcessorMessage.RecallCommand.RegisterID);   
    printf(",  Register Value = ");
    dec(CoProcessorMessage.RecallCommand.RegisterValue);
	DoubleToAscii(CoProcessorMessage.RecallCommand.RegisterValue, af1);
   	puts(af1);
    printf("\r\n\r\n");
#endif	
													 
	// Send the command message to the SPI Slave Controller
	SendSPIMessage();
	
	// Receive and process the response message from the SPI Slave Controller
	ReceiveSPIMessage();
	
	// Give enough time for the command to complete
	
	pause(RecallCommandDelay);
	//?????????????????????????
		
 	// Check the status of the the current controller to see if the command has been 
	// executed correctly.
			
	//while (CheckControllerState() != DONE);

	return RecallCommandStatus;
}


//**********************************************************************************
//* SendResetCommand - Send the Reset the SPI Slave Controller.  This command 
//* resets the selected controller.  It may be used to implement an emergency stop 
//* function using the PIC's watchdog timer. 
//**********************************************************************************
byte SendResetCommand(void)
{ 

	byte ResetCommandStatus;
 	word ResetCommandDelay = 2;											// Delay to allow the duty command to complete
	
	// Clear the SPI send and receive buffers
	//FlushSPIBuffers();

	// Build Reset Command Message
	
	CoProcessorMessage.ResetCommand.MessageID = MessageIndex++;	// Message ID range (0000 - FFFF) Hex
	CoProcessorMessage.ResetCommand.Command = RESET_COMMAND;		// Send Reset Command
   
#ifdef TEST
    printf("Master: Send Reset Command...");
  	printf("MessageID = ");
   	dec(CoProcessorMessage.ResetCommand.MessageID); 
   	printf("\r\n\r\n");
#endif 


	// Send the command message to the SPI Slave Controller
	SendSPIMessage();
	
	// Receive and process the response message from the SPI Slave Controller
	ReceiveSPIMessage();

	// Give enough time for the command to complete
	
	pause(ResetCommandDelay);
	//?????????????????????????????
	
 	// Check the status of the the current controller to see if the command has been 
	// executed correctly.
			
	//while (CheckControllerState() != DONE);
 
	return ResetCommandStatus;
	
}

//**********************************************************************************
//* SendGetStatusCommand - Request to get the latest status from the selected Slave*
//* controller.  The message includes the controller state, any errors detected and*
//* any diagnostic messages reported.                                              *
//**********************************************************************************
byte SendGetStatusCommand(void)
{ 

	byte GetStatusCommandStatus;
 	word GetStatusCommandDelay = 5;											// Delay to allow the duty command to complete
	
	// Clear the SPI send and receive buffers
	//FlushSPIBuffers();

	// Build Reset Command Message
	
	CoProcessorMessage.GetStatusCommand.MessageID = MessageIndex++;	// Message ID range (0000 - FFFF) Hex
	CoProcessorMessage.GetStatusCommand.Command = GET_STATUS_COMMAND;		// Send Get Status Command
   
#ifdef TEST
    printf("Master: Send Get Status Command...");
  	printf("MessageID = ");
   	dec(CoProcessorMessage.GetStatusCommand.MessageID);
  
   	printf("\r\n\r\n");
#endif 

	// Send the command message to the SPI Slave Controller
	SendSPIMessage();
	
	// Receive and process the response message from the SPI Slave Controller
	ReceiveSPIMessage();
	
	// Give enough time for the command to complete
	
	pause(GetStatusCommandDelay);
	//?????????????????????????????
		
 	// Check the status of the the current controller to see if the command has been 
	// executed correctly.
			
	//while (CheckControllerState() != DONE);
 	

	return GetStatusCommandStatus;
	
}
		
//**********************************************************************************
//**********************************************************************************
//**********************************************************************************
//**********************************************************************************
// Functions used to process responses from the Slave Math Co-processor  

// Receive Responses from Slave here...

//**********************************************************************************
//* ReceiveCalculateCommand - Evalueates the RPN expressions that are stored in    *
//* the message expression buffer.  Multiple expressions can be sent to the        *
//* co-processor for evaluation, if they are separated by a ";" semi-colon.  Large *
//* expressions can span multiple messages to the co-processor.                    *
//**********************************************************************************
byte ReceiveCalculateCommand(void)  
{
 	byte ReceiveCalculateCommandStatus = TRUE;							// Status return
 	byte i;
 	
 	// Parse and Evalueate the RPN expression contained in the co-processor message buffer ...
 	
 	
 	
#ifdef TEST
    printf("slave: Calculate Command:");
  	printf("  MessageID = ");
   	dec(CoProcessorMessage.CalculateCommand.MessageID);
   	printf("\r\n");
    printf("Buffer = !");
   	for (i=0; i<20;i++)
   	{
   		printf("i = ");
   		dec(i);
   		printf("Buffer = ");
   		dec(CoProcessorMessage.CalculateCommand.Buffer[i]);
   		printf("\r\n");
   	}

    printf("!");
   	printf("\r\n\r\n");
#endif

 				
	return ReceiveCalculateCommandStatus;
}

//**********************************************************************************
//* ReceiveGetAnswers - Get the results from the latest arithmetic operations      *
//* carried out during the last cycle. Up to 7 separate single precision floating  *
//* point or 3 double precision  values may be returned during a single message    *
//* transfer.                                                                      *
//**********************************************************************************
byte ReceiveGetAnswers(void)
{
 	byte ReceiveGetAnswersStatus = TRUE;								// Status return
   	byte AnswerID;									 		    // Local Answer Loop Index 
    				
  	// Send  latest answers from requested RPN expressions back to the master.

#ifndef TEST
    printf("slave: Get Answers Command: \r\n");
  	printf("  MessageID = ");
   	dec(CoProcessorMessage.GetAnswersCommand.MessageID);
    printf(", Number Of Answers = ");
    dec(CoProcessorMessage.GetAnswersCommand.NumberOfAnswers);  
    printf("\r\n");
#endif

	// Display all the answers in floating point format
	
 	for (AnswerID = 0; AnswerID < CoProcessorMessage.GetAnswersCommand.NumberOfAnswers; AnswerID++)
	{	
		// Copy local answers to Answer message
		Answers[AnswerID] = CoProcessorMessage.GetAnswersCommand.Answers[AnswerID];
#ifndef TEST
			
  		printf("  Answer ID = ");
    	dec(AnswerID);
    	printf(", Answer = ");
		//DoubleToAscii(CoProcessorMessage.GetAnswersCommand.Answers[AnswerID], af1);
   		//puts(af1);	      		DoubleToAscii(CoProcessorMessage.GetAnswersCommand.Answers[AnswerID], af1);
   		hex4(CoProcessorMessage.GetAnswersCommand.Answers[AnswerID]);	              	
    	printf("\r\n");
#endif

	}
		 
	return ReceiveGetAnswersStatus;
}
			
//**********************************************************************************
//* ReceiveClearMemoryCommand - Erases the contents of the Math co-processor's 
//* floating point storage registers. These same registers are used for storing and
//* retrieving temporary values and answers. It does this by setting floating point 
//* registers 0..MAX_FLOAT_REGISTERS to zero.  Registers 0..10 are reserved for answers.
//**********************************************************************************
byte ReceiveClearMemoryCommand(void)
{
 	byte ReceiveClearMemoryCommandStatus = TRUE;							// Status return
    byte RegisterID;									 		    // Local Loop Index
	
#ifndef TEST

 	// Display the current register values   
 
  	printf("  MessageID = ");
   	dec(CoProcessorMessage.CLearMemoryCommand.MessageID);
    printf("save: Clear Memory Command Completed.");
   	printf("  MessageID = ");
   	dec(CoProcessorMessage.CLearMemoryCommand.MessageID);
   printf("\r\n\r\n");
#endif

	return ReceiveClearMemoryCommandStatus;
}

//**********************************************************************************
//* ReceiveStoreCommand - Stores a floating point value to a selected floating     *
//* point register.     The register ID range is 0..MAX_FLOAT_REGISTERS.           *
//**********************************************************************************
byte ReceiveStoreCommand(void)
{
 	byte ReceiveStoreCommandStatus = TRUE;										// Status return
			
 	// Store the value into the selected register.
  					
 	Registers[CoProcessorMessage.StoreCommand.RegisterID] = CoProcessorMessage.StoreCommand.RegisterValue;
  	 
#ifndef TEST
    printf("slave: Store Command:");
  	printf("  MessageID = ");
   	dec(CoProcessorMessage.GetStatusCommand.MessageID);
   	printf("  Register ID = ");
   	dec(CoProcessorMessage.StoreCommand.RegisterID);
    printf("Register Value = ");
    hex4(Registers[CoProcessorMessage.StoreCommand.RegisterID]);
    printf("\r\n\r\n");
#endif 

	return ReceiveStoreCommandStatus;
}
  					
//**********************************************************************************
//* ReceiveRecallCommand - Retrieves a floating point value from the selected      *
//* floating point register. The register ID range is 0..MAX_FLOAT_REGISTERS.      *                                                               *
//**********************************************************************************
byte ReceiveRecallCommand(void)
{
 	byte ReceiveRecallCommandStatus = TRUE;									// Status return
 	
    // Accept number of samples xxxx to be used for calibration and data
   	// collection used for reading the DC Motor Control Pots, where the 
   	// number of samples range 0000..FFFF.
	
  	CoProcessorMessage.RecallCommand.RegisterValue = Registers[CoProcessorMessage.RecallCommand.RegisterID];
  	
#ifndef TEST
    printf("slave: Recall Command:");
  	printf("  MessageID = ");
   	dec(CoProcessorMessage.RecallCommand.MessageID);
   	printf("  Register ID = ");
   	dec(CoProcessorMessage.RecallCommand.RegisterID);
    printf("Register Value = ");
    hex4(CoProcessorMessage.RecallCommand.RegisterValue);
    printf("\r\n\r\n");
#endif 

	return ReceiveRecallCommandStatus;
}



 

//**********************************************************************************
//* ReceiveResetResponse - Process the Reset the SPI Slave Controller.  This command 
//* resets the selected controller.  It may be used to implement an emergency stop 
//* function using the PIC's watchdog timer. 
//**********************************************************************************
byte ReceiveResetResponse(void)
{
	byte ResetCommandStatus;
	
#ifndef TEST
    printf("Master: Received Reset Command Response:  ");
  	printf("MessageID = ");
   	dec(CoProcessorMessage.ResetCommand.MessageID); 
   	printf("\r\n\r\n");
#endif

	// Reset the Controller by allowing the Watchdog Timer to timeout.
	
	return ResetCommandStatus;

}

//**********************************************************************************
//* ReceiveGetStatusResponse - Get the latest status from the selected Slave       *
//* controller.  The message includes the controller state, any errors detected and*
//* any diagnostic messages reported.   The current state is returned from the     *
//* function call, while the error and diagnostics codes are saved globally.       *
//**********************************************************************************
byte ReceiveGetStatusResponse(void)
{ 	
	byte GetStatusCommantStatus;
 	
	//CoProcessorMessage.GetStatusCommand.MessageID = MessageIndex++;			     			// Message ID range (0000 - FFFF) Hex
	//CoProcessorMessage.GetStatusCommand.Command = GET_STATUS_COMMAND;		     			// Send Get Status Command
    ControllerState = CoProcessorMessage.GetStatusCommand.ControllerState; 		 			// Get current Controller State
    ControllerErrorCode = CoProcessorMessage.GetStatusCommand.ControllerErrorCode; 			// Get current Controller Error Code
    ControllerDiagnosticCode = CoProcessorMessage.GetStatusCommand.ControllerDiagnosticCode;  // Get current Diagnostic Code
    
#ifdef TEST
    printf("Master: Received Get Status Command Response:  ");
  	printf("MessageID = ");
   	dec(CoProcessorMessage.GetStatusCommand.MessageID);
   	printf("  ControllerState = ");
   	dec(ControllerState);
   	printf("  ControllerErrorCode = ");
   	dec(ControllerErrorCode);
   	printf("  ControllerDiagnosticCode = ");
   	dec(ControllerDiagnosticCode);
   
   	printf("\r\n\r\n");
#endif 
 
	return GetStatusCommantStatus;
	
}

#ifdef PARALLEL_PORT_INTERFACE
//******************************************************************************
//* InitializeParallelPort - Initializes the 8-Bit Parallel Slave Port.
//******************************************************************************
void InitializeParallelPort(void)
{
	int i;
	OutData = 0;					// clrf	OUTDATA		;Clear data registers
	InData = 0;						// clrf	INDATA
									// bsf	STATUS,RP0	;Select register page 1
 
   	// configure port A for all digital inputs


    //OpenADC(ADC_FOSC_32 &
    //        ADC_RIGHT_JUST & ADC_0ANA_0REF, ADC_CH0 & ADC_CH1 & ADC_CH2 & ADC_CH3 & ADC_CH4 & ADC_CH5 & ADC_CH6 & ADC_CH7 & ADC_INT_OFF);
    //OpenADC(ADC_0ANA_0REF, ADC_INT_OFF);

    // Configure PORTB interrupt with Interrupt on change off and pull-up resistors disabled

    //OpenPORTB(PORTB_CHANGE_INT_OFF & PORTB_PULLUPS_OFF);
    
	EnablePullups();				// Enable PORTB pullups (inputs)

  
    // Initialize ports

    //PORTA = 0x00;                                         // Clear PORTA
    PORTB = 0x00;                                         	// Clear PORTB
    //PORTC = 0x00;                                         // Clear PORTC. Some bits of PORTC are used by the USART.
    PORTD = 0x00;                                          	// Clear PORTD
    PORTE = 0x00;                                          	// Clear PORTE

    // Initialize port I/O direction

    //TRISA = 0xFF;                                         // PORT A all inputs.
    TRISB = 0x00;                                         	// PORT B all outputs.
    //TRISC = 0x00;                                         // PORT C all outputs. Some bits of PORTC are used by the USART.
    TRISD = 0x00;                                          	// PORT D all outputs (switches select each pin state)
  
	// Configure PORT E for 8 Bit Slave Parallel Port as outputs for Master

  
 	BUSY_DIR = INPUT;				   						// Set Slave BUSY Status Pin Direction  to input

 	RD_BAR_DIR	= OUTPUT;       							// RD_BAR Control Pin Direction
 	WR_BAR_DIR 	= OUTPUT;       							// WR_BAR Control Pin Direction 
 	CS_BAR_DIR 	= OUTPUT;       							// CS_BAR Control Pin Direction 

 
	// Initialize Message buffers

 
	// Initialization Logic for SPI Slave Buffer Interface goes here ...
	for (i=0; i<SPI_BUFFER_SIZE; i++)
	{
 		SPI_Rx_Buffer[i] = 0;                  	// Clear SPI Receive Buffer 
 		SPI_Tx_Buffer[i] = 0;                  	// Clear SPI Transmit Buffer   
	}
	
	
	for (i=0; i<MESSAGE_BUFFER_SIZE; i++)
	{
 		CoProcessorMessage.CalculateCommand.Buffer[i] = 0;            			// Clear the message  
 	}
									 
}

#endif		 	
 
