//********************************************************************************
//*         Name:  coprocm.h                                                     *               
//*                                                                              *
//*       Author:  Daniel F. Ramirez                                             *
//*                                                                              *
//*          Date:  06/13/2003                                                   *
//*                                                                              *
//*   Description:                                                               *
//*                                                                              *
//*                                                                              *
//*     This is a Master Math Co-Processor application that uses a PIC18F8720 to * 
//* send expressions to the embedded Math Co-Processor using the USART interface,*
//* in order to compute floating point arithmetic expressions in RPN format.  It *
//* can be used for Sumo Robots that require floating point for navigation and   *
//* tracking of an oponent with minimal space, power and weight requirements are *
//* prime candidates for using this co-processor, since it is very compact.      *
//* The master can be a PIC or STAMP or any other micro-controller that requires *
//* a math co-processor and has at least 4 spare I/O pins. In addition to the    *
//* basic arithmetic functions, the math co-processor can also evaluate          *
//* expressions that contain trigonometric and scientific functions.             *
//*                                                                              *
//*      A communication interface is provided that allows the host to use high  *
//* speed asynchronous serial link or SPI or I2C Master to request floating point*
//* equation calculations or to pass it individual sensor readings, average      *
//* sensor readings, calibrating a sensor or to perform a sensor processing      *
//* algorithm using command message packets.  The commands available are         *
//* described below.  This header file provides the common Master and Slave      *
//* message formats and descriptors.                                             *
//*/*                                                                            *
//* The high priority ISR will invoke the "Z;" reset command via the serial      *
//* interface. Both the serial interface and the SPI interface are supported by  *
//* defining the approprate compiler flag.                                       *
//*                                                                              *
//*                                                                              *
//*     This master controller can be used to communicate with a PC host or any  *
//* USART, I2C or SPI slave, in order to  give the microcontroller the capability*
//* to serve as an SPI or I2C slave controller in order to provide it the        *
//* floating point coprocessor functions.                                        *  
//* It can also be used for robotic applications where multiple controllers and  *
//* SPI devices are used, with a minimum of data and control lines needed to     *
//* connect them.  Each I2C device has its own unique address and only listens   *
//* to commands associated with it.                                              *
//*                                                                              *
//*     Future upgrades to this application will include implementing various    *
//* filters including FIR, FFR, Kalman and Median filters.  Both FFT and IFFT    *
//* functions and a Histogram function are also in the works. Advanced algorithms*
//* such as Vector Force Fields (VFF), Fuzzy Logic and Neural Networks are new   *
//* areas that may be explored using this application.   In addition, this       *
//* application may be able to take advantage of the new DSPic Signal Processors *
//* (DSP's) become available from Microchip.                                     *
//*                                                                              *
//* Usage:                                                                       *
//*	The following Math Co-Processor Controller command messages are available:   *
//*                                                                              * 
//* calculate    - Evalueates the RPN expressions that are stored in the         *
//* message expression buffer.  Multiple expressions can be sent to the          *
//* co-processor for evaluation, if they are separated by a ";" semi-colon.      *
//* Large expressions can span multiple messages to the co-processor.            *
//*                                                                              * 
//* get_answers  - Get the results from the latest arithmetic operations         *
//* carried out during the last cycle. Up to 7 separate single precision floating*
//* point and 3 double precision  values may be returned during a single message *
//* transfer.                                                                    *
//*                                                                              *
//* clear_memory - Erases the contents of the Math co-processor's storage        * 
//* registers. These same registers are used for storing and retrieving temporary*
//* values. It clears floating point registers 0..MAX_FLOAT_REGISTERS.           *
//*                                                                              *
//* store   - Stores a floating point value to a selected floating point         *
//* register. The register ID range is 0..MAX_FLOAT_REGISTERS.                   *                                                            
//*                                                                              *
//* recall 	- Retrieves a floating point value from the selected floating        *
//* point register. The register ID range is 0..MAX_FLOAT_REGISTERS.             *                                                         
//*                                                                              *                                                                     
//*	GetStatus - Get the Math Co-processor status including current state, and    *  
//* any error conditions encountered i.e. division by 0, overflow, underflow.    *
//*                                                                              *
//* ResetCommand - Reset the DC motor Controller (brake the motor).  This        *
//* command can be used to implement an emergency stop function.                 *  
//*                                                                              *
//*                                                                              * 
//*                                                                              *
//********************************************************************************
    
 
#ifndef COPROCM_LOADED  /* prevent multiple loading */
#define COPROCM_LOADED

 
#include "flags.h"                                    // Global flags used for MPSIM debugging 
#include "stdio.h"                                    // I/O support functions  
#include "network.h"                                  	// Include support for networking
//#include "dcmotor.h"                                  	// Include common messages

// Functions to send the coprocessor messages and return results

byte GetResults(float *Results);
   
// Functions used to send commands to the slave DC Motor Controller
byte SendCalculateCommand(const rom char *Data);
byte SendGetAnswers(void);
byte SendClearMemoryCommand(void);
byte SendStoreCommand(void);
byte SendRecallCommand(void);
byte SendResetCommand(void);
byte SendGetStatusCommand(void);
 
// Functions used to process responses from the slave Math Co-Processor

byte ReceiveCalculateCommand(void);  
byte ReceiveGetAnswers(void);
byte ReceiveClearMemoryCommand(void);
byte ReceiveStoreCommand(void);
byte ReceiveRecallCommand(void);
#ifdef NEW_CODE
byte ReceiveFilterCommand(void);
byte ReceiveFFTCommand(void);
byte ReceiveIFFTCommand(void);
byte ReceiveKalmanFilterCommand(void);
byte ReceiveMedianFilterCommand(void);
byte ReceiveHistogramCommand(void);
#endif


byte ReceiveGetStatusResponse(void);
byte ReceiveResetResponse(void);

//**********************************************************************************
//* SendSPIMessage - Sends a command message to the SPI Slave and stores the data received 
//* into the SPI_Rx_Buffer so that it may be processed by Reveive Message. 
//**********************************************************************************
byte SendSPIMessage(void);

//**********************************************************************************
//* ReceiveSPIMessage - Receives and processes the message that is in the SPI_Rx_Buffer 
//* which was simultaneously sent back from the SPI Slave Controller.  The message may
//* contain data or status information.
//**********************************************************************************
byte ReceiveSPIMessage(void);

// High Level Master DC Motor Controller functions

void Neutral(void);									// Motor is in Neutral or Stopped
void Neutral_To_Forward(void);						// Accelerate from Neutral to maximum Forward  Speed
void Forward_To_Neutral(void);						// De-accelerate from maximum Forward speed to Neutral
void Neutral_To_Reverse(void);						// Accelerate from Neutral to maximum Reverse speed
void Reverse_To_Neutral(void);						// De-accelerate from maximum Reverse speed to Neutral
void Left(void);									// Turn left
void HardLeft(void);								// Turn hard left
void Right(void);									// Turn right
void HardRight(void);								// Turn hard right


#endif

